from storages.backends.s3boto3 import S3Boto3Storage


class S3OptimizedUploadStorage(S3Boto3Storage):
    """
    Class for an optimized S3 storage.

    This storage prevents unnecessary operation to copy with the general ``upload_fileobj``
    command when the object already is a S3 object where the faster copy command can be used.

    The assumption is that ``content`` contains a S3 object from which we can copy.

    See also discussion here: https://github.com/codingjoe/django-s3file/discussions/126
    """

    def _save(self, name, content):
        # Basically copy the implementation of _save of S3Boto3Storage
        # and replace the obj.upload_fileobj with a copy function
        cleaned_name = self._clean_name(name)
        name = self._normalize_name(cleaned_name)
        params = self._get_write_parameters(name, content)

        if (
            self.gzip
            and params["ContentType"] in self.gzip_content_types
            and "ContentEncoding" not in params
        ):
            content = self._compress_content(content)
            params["ContentEncoding"] = "gzip"

        obj = self.bucket.Object(name)
        # content.seek(0, os.SEEK_SET)  # Disable unnecessary seek operation
        # obj.upload_fileobj(content, ExtraArgs=params)  # Disable upload function

        if not hasattr(content, "obj") or not hasattr(content.obj, "key"):
            raise TypeError(
                "The content object must be a S3 object and contain a valid key."
            )

        # Copy the file instead uf uploading
        obj.copy({"Bucket": self.bucket.name, "Key": content.obj.key}, ExtraArgs=params)

        return cleaned_name



from django.apps import AppConfig
from django.core import checks

from .checks import storage_check


class S3FileConfig(AppConfig):
    name = "s3file"
    verbose_name = "S3File"

    def ready(self):
        from django import forms
        from django.core.files.storage import FileSystemStorage, default_storage
        from storages.backends.s3boto3 import S3Boto3Storage

        from .forms import S3FileInputMixin

        if (
            isinstance(default_storage, (S3Boto3Storage, FileSystemStorage))
            and S3FileInputMixin not in forms.ClearableFileInput.__bases__
        ):
            forms.ClearableFileInput.__bases__ = (
                S3FileInputMixin,
            ) + forms.ClearableFileInput.__bases__

        elif S3FileInputMixin in forms.ClearableFileInput.__bases__:
            forms.ClearableFileInput.__bases__ = tuple(
                cls
                for cls in forms.ClearableFileInput.__bases__
                if cls is not S3FileInputMixin
            )

        checks.register(storage_check, checks.Tags.security, deploy=True)



import base64
import logging
import pathlib
import uuid

from django.conf import settings
from django.utils.functional import cached_property
from storages.utils import safe_join

from s3file.middleware import S3FileMiddleware
from s3file.storages import get_aws_location, storage

logger = logging.getLogger("s3file")


class S3FileInputMixin:
    """FileInput that uses JavaScript to directly upload to Amazon S3."""

    needs_multipart_form = False
    upload_path = safe_join(
        str(get_aws_location()),
        str(
            getattr(
                settings, "S3FILE_UPLOAD_PATH", pathlib.PurePosixPath("tmp", "s3file")
            )
        ),
    )
    expires = settings.SESSION_COOKIE_AGE

    @property
    def bucket_name(self):
        return storage.bucket.name

    @property
    def client(self):
        return storage.connection.meta.client

    def build_attrs(self, *args, **kwargs):
        attrs = super().build_attrs(*args, **kwargs)

        accept = attrs.get("accept")
        response = self.client.generate_presigned_post(
            self.bucket_name,
            str(pathlib.PurePosixPath(self.upload_folder, "${filename}")),
            Conditions=self.get_conditions(accept),
            ExpiresIn=self.expires,
        )

        defaults = {
            "data-fields-%s" % key: value for key, value in response["fields"].items()
        }
        defaults["data-url"] = response["url"]
        # we sign upload location, and will only accept files within the same folder
        defaults["data-s3f-signature"] = S3FileMiddleware.sign_s3_key_prefix(
            self.upload_folder
        )
        defaults.update(attrs)

        try:
            defaults["class"] += " s3file"
        except KeyError:
            defaults["class"] = "s3file"
        return defaults

    def get_conditions(self, accept):
        conditions = [
            {"bucket": self.bucket_name},
            ["starts-with", "$key", str(self.upload_folder)],
            {"success_action_status": "201"},
        ]
        if accept and "," not in accept:
            top_type, sub_type = accept.split("/", 1)
            if sub_type == "*":
                conditions.append(["starts-with", "$Content-Type", "%s/" % top_type])
            else:
                conditions.append({"Content-Type": accept})
        else:
            conditions.append(["starts-with", "$Content-Type", ""])

        return conditions

    @cached_property
    def upload_folder(self):
        return str(
            pathlib.PurePosixPath(
                self.upload_path,
                base64.urlsafe_b64encode(uuid.uuid4().bytes)
                .decode("utf-8")
                .rstrip("=\n"),
            )
        )  # S3 uses POSIX paths

    class Media:
        js = ("s3file/js/s3file.js" if settings.DEBUG else "s3file/js/s3file.min.js",)



import logging
import pathlib

from django.core import signing
from django.core.exceptions import PermissionDenied, SuspiciousFileOperation
from django.utils.crypto import constant_time_compare

from . import views
from .storages import get_aws_location, local_dev, storage

logger = logging.getLogger("s3file")


class S3FileMiddleware:
    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        file_fields = request.POST.getlist("s3file")
        for field_name in file_fields:

            paths = request.POST.getlist(field_name)
            if paths:
                try:
                    signature = request.POST[f"{field_name}-s3f-signature"]
                except KeyError:
                    raise PermissionDenied("No signature provided.")
                try:
                    request.FILES.setlist(
                        field_name, list(self.get_files_from_storage(paths, signature))
                    )
                except SuspiciousFileOperation as e:
                    raise PermissionDenied("Illegal file name!") from e

        if local_dev and request.path == "/__s3_mock__/":
            return views.S3MockView.as_view()(request)

        return self.get_response(request)

    @classmethod
    def get_files_from_storage(cls, paths, signature):
        """Return S3 file where the name does not include the path."""
        location = get_aws_location()
        for path in paths:
            path = pathlib.PurePosixPath(path)
            if not constant_time_compare(
                cls.sign_s3_key_prefix(path.parent), signature
            ):
                raise PermissionDenied("Illegal signature!")
            try:
                relative_path = str(path.relative_to(location))
            except ValueError as e:
                raise SuspiciousFileOperation(
                    f"Path is outside the storage location: {path}"
                ) from e

            try:
                f = storage.open(relative_path)
                f.name = path.name
                yield f
            except (OSError, ValueError):
                logger.exception("File not found: %s", path)

    @classmethod
    def sign_s3_key_prefix(cls, path):
        """
        Signature to validate the S3 keys passed the middleware before fetching files.

        Return a base64-encoded HMAC-SHA256 of the upload folder aka the S3 key-prefix.
        """
        return signing.Signer(salt="s3file.middleware.S3FileMiddleware").signature(path)



"""A lightweight file uploader input for Django and Amazon S3."""

import django

from . import _version

__version__ = _version.version
VERSION = _version.version_tuple

if django.VERSION < (4, 0):
    default_app_config = "s3file.apps.S3FileConfig"
